/* Copyright 2004 NetApp, Inc., contribution by Yen-Chau Ngo.
 * Copyright 2008 NetApp, Inc., contribution by Ganesh R Narayanan.
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 * License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License v2 for more
 * details.
 *
 * This driver provides services for the Agent and LM77 interrupts.
 */ 

#include <linux/devfs_fs_kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/init.h>
#include <asm/ibm4xx.h>
#include <asm/irq.h>
#include <linux/ibm_ocp_gpio.h>

#include "agentdrv.h"


#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


#ifndef THIS_MODULE
#define THIS_MODULE NULL
#endif


#ifdef MODULE
extern int init_module(void);
extern int cleanup_module(void);
#endif /* MODULE */


#define DCRN_UIC0_SR 0xC0
#define DCRN_UIC0_ER 0xC2
#define DCRN_UIC0_CR 0xC3
#define DCRN_UIC0_PR 0xC4
#define DCRN_UIC0_TR 0xC5
#define DCRN_UIC0_MSR 0xC6

#define IRQ_MAX_TRY 10

#define DBG_LEVEL 1

#if DBG_LEVEL > 0
#  define DBG(x...)	printk( KERN_DEBUG x )
#else
#  define DBG(x...)	((void)0)
#endif

#if DBG_LEVEL > 1
#  define DBG2(x...) 	printk( KERN_DEBUG x )
#else
#  define DBG2(x...) 	((void)0)
#endif


int __init agentdrv_init(void);
static int __init agentdrv_cleanup(void);
static int agentdrv_ioctl (struct inode *inode, struct file *file, 
			   unsigned int cmd, unsigned long arg);
static ssize_t agentdrv_read (struct file *file, char *buf, size_t count, 
			      loff_t *offset);
static int agentdrv_open (struct inode *inode, struct file *file);
static int agentdrv_release (struct inode *inode, struct file *file);

static void agentdrv_ISR (int irq, void *dev_id, struct pt_regs *regs);

static struct file_operations agentdrv_fops = {
	owner:		THIS_MODULE,
	llseek:		no_llseek,
	read:		agentdrv_read,
	write:		NULL,
	ioctl:		agentdrv_ioctl,
	open:		agentdrv_open,
	release:	agentdrv_release,
};


static int __initdata agentdrv_initialized = 0;
static volatile char agentdrv_irq_source=0;
static wait_queue_head_t  waitq;
static int agentdrv_major=0;
static devfs_handle_t devfs_handle = NULL;
static int agentdrv_irqs = 0x00; /* bit fields defined as in UIC_SR */
static spinlock_t agentdrv_lock = SPIN_LOCK_UNLOCKED;


/* Interrupt Service Routine */
void agentdrv_ISR (int irq, void *dev_id, struct pt_regs *regs)
{
	disable_irq(irq);
	DBG2("\nIRQ %d asserted!!!\n", irq);

	if (irq == AGENT_IIRQ){
		/* disable all other IRQs */
		disable_irq(AGENT_IRQ);
		disable_irq(LM77_IRQ);
		disable_irq(LM77_CIRQ);
		agentdrv_irq_source = (1 << (31 - irq));
	}
	else
		agentdrv_irq_source |= (1 << (31 - irq));

	/* Wake up any blocked readers */
	if (waitqueue_active(&waitq)) {
		wake_up_interruptible(&waitq);
	}

}

/* Request an irq.
   Return 0 if successful, non-zero otherwise.
 */
static int agentdrv_request_irq(int irq)
{
	int res=0;

	res = request_irq(irq, agentdrv_ISR, SA_INTERRUPT, 
			  AGENTDRV_DEVNAME, NULL);
	if (res != 0){
		printk(KERN_ERR "agentdrv ERROR: failed to request IRQ %d\n", irq);
	}
	else{
		agentdrv_irqs |= (1 << (31 -irq));
		DBG("agentdrv: IRQ %d registered\n", irq);
	}
	return res;
}


/* Release an irq */
static int agentdrv_release_irq(int irq)
{
	free_irq(irq, NULL);
	DBG("agentdrv: Released IRQ %d\n", irq);

	return 0;
}

/* Register all irq's that will be handled by agentdrv_ISR */
static int agentdrv_register_irqs(void)
{
	/* setting level trigger for all supported IRQs */
	int uic = mfdcr(DCRN_UIC0_TR);
	DBG("TR default: 0x%08x\n", uic);
	mtdcr(DCRN_UIC0_TR, uic & 0xffffff87);

	/* check uic conf */
	DBG("SR: 0x%08x   ER: 0x%08x\n", 
	    mfdcr(DCRN_UIC0_SR), mfdcr(DCRN_UIC0_ER));
	DBG("CR: 0x%08x   PR: 0x%08x\n", 
	    mfdcr(DCRN_UIC0_CR), mfdcr(DCRN_UIC0_PR));
	DBG("TR: 0x%08x   MSR: 0x%08x\n", 
	    mfdcr(DCRN_UIC0_TR), mfdcr(DCRN_UIC0_MSR));


	agentdrv_request_irq(AGENT_IRQ);		
	agentdrv_request_irq(AGENT_IIRQ);
	agentdrv_request_irq(LM77_CIRQ);
	agentdrv_request_irq(LM77_IRQ);

	DBG("ER: 0x%08x\n", mfdcr(DCRN_UIC0_ER));
	return 0;
}	


/* Unregister all irq's that have been handled by agentdrv_ISR */
static int agentdrv_unregister_irqs(void)
{
	agentdrv_release_irq(AGENT_IRQ);
	agentdrv_release_irq(AGENT_IIRQ);
	agentdrv_release_irq(LM77_IRQ);
	agentdrv_release_irq(LM77_CIRQ);
	return 0;
}






/* Blocking read.
   Return number of bytes read. 
   If successful the 1 byte read is the latest IRQ# that got asserted.
 */
ssize_t agentdrv_read (struct file *file, char *buf, size_t count, 
                            loff_t *offset)
{
	unsigned long flags;

	DBG2("agentdrv_read\n");

	if (!buf || count < 1)
		return 0;

	if (agentdrv_irq_source == 0){
		interruptible_sleep_on(&waitq);

		if (signal_pending(current)) {
			return -ERESTARTSYS;
		}

	}
	buf[0] = agentdrv_irq_source;

	spin_lock_irqsave(agentdrv_lock, flags);
	agentdrv_irq_source &= ~(buf[0]);
	spin_unlock_irqrestore(agentdrv_lock, flags);

	DBG2("buf = 0x%02x, source = 0x%02x\n", buf[0], agentdrv_irq_source);

	return 1;
}


int agentdrv_ioctl (struct inode *inode, struct file *file, unsigned int cmd, 
		    unsigned long arg)
{
	volatile int uic;
	int stat, count=0;

	DBG2("agentdrv_ioctl i2c-%d ioctl, cmd: 0x%x, arg: %lx.\n", 
	    MINOR(inode->i_rdev),cmd, arg);

	switch (cmd){
	case AGENTDRV_SET_GPIO:
		ibm_gpio_kern_set(arg);
		break;
	case AGENTDRV_CLR_GPIO:
		ibm_gpio_kern_clr(arg);
		break;

	case AGENTDRV_ENABLE_IRQ:
		/* Enable an IRQ if it's not currently enabled. */
		stat = 1 << (31 - arg);
		uic = mfdcr(DCRN_UIC0_ER);
		DBG2("ER = 0x%08x, stat = 0x%08x\n", uic, stat);
		while (((uic & stat) == 0) && (count++ < IRQ_MAX_TRY)){
			DBG2("\nenable_irq %ld\n", arg);
			enable_irq(arg);
			uic = mfdcr(DCRN_UIC0_ER);
		}
		if ((uic & stat) == 0)
			return -1;
		break;
	case AGENTDRV_DISABLE_IRQ:
		/* Disable an IRQ if it's currently enabled. */
		stat = 1 << (31 - arg);
		uic = mfdcr(DCRN_UIC0_ER);
		DBG2("ER = 0x%08x, stat = 0x%08x\n", uic, stat);
		while (((uic & stat) != 0) && (count++ < IRQ_MAX_TRY)){
			disable_irq(arg);
			DBG2("disable_irq %ld\n", arg);
			uic = mfdcr(DCRN_UIC0_ER);
		}
		if ((uic & stat) != 0)
			return -1;
		break;
	case AGENTDRV_SUPPORTED_IRQS:
		/* Return supported IRQs mask that have been 
		   successfully registered.
		*/
		return agentdrv_irqs;
	case AGENTDRV_UIC_ER:
		return mfdcr(DCRN_UIC0_ER);

	case AGENTDRV_UIC_STATUS:
		/* freshen the intr bits
		   and return the current status.
		*/
		uic = mfdcr(DCRN_UIC0_SR);
		DBG2("before clear uic: 0x%08x\n", uic);
		mtdcr(DCRN_UIC0_SR, uic | AGENTDRV_UIC_MASK);
		uic = mfdcr(DCRN_UIC0_SR);
		DBG2("after clear uic: 0x%08x\n", uic);
		return mfdcr(DCRN_UIC0_SR);
	default:
		printk(KERN_ERR "agentdrv ERROR: invalid ioctl 0x%02x\n", cmd);
		return -EINVAL;
	}
	return 0;
}


int agentdrv_open (struct inode *inode, struct file *file)
{
	DBG2("agentdrv_open\n");

	return 0;
}

static int agentdrv_release (struct inode *inode, struct file *file)
{
	DBG2("agentdrv_release\n");

	return 0;
}



int __init agentdrv_init(void)
{
	char name[10];

	DBG2("agentdrv_init\n");
	agentdrv_initialized = 0;

	agentdrv_major = devfs_register_chrdev(0, AGENTDRV_DEVNAME, 
					       &agentdrv_fops);
	if (!agentdrv_major){
		printk(KERN_ERR "agentdrv: error getting major device\n");
		return -EIO;
	}
	printk(KERN_INFO "\nagentdrv.o: major %d\n", agentdrv_major);

	agentdrv_initialized++;

	sprintf(name, "%s", AGENTDRV_DEVNAME);
	devfs_register (devfs_handle, name,
			DEVFS_FL_DEFAULT, agentdrv_major, 0,
			S_IFCHR | S_IRUSR | S_IWUSR,
			&agentdrv_fops, NULL);



	/* Init wait queue for blocking readers */
	init_waitqueue_head(&waitq);


	/* Register IRQs, which will be handled by this driver */
	agentdrv_register_irqs();
	return 0;
}

int __init agentdrv_cleanup(void)
{
	int res;
	DBG("agentdrv_cleanup\n");

	agentdrv_unregister_irqs();

	if (agentdrv_initialized >= 1) {
		devfs_unregister(devfs_handle);
		res = devfs_unregister_chrdev(agentdrv_major, AGENTDRV_DEVNAME);

		if (res){
			printk(KERN_ERR "agentdrv: unable to release major %d\n",
			       agentdrv_major);
			return res;
		}
		agentdrv_initialized--;
	}
	
	return 0;
}

EXPORT_NO_SYMBOLS;

#ifdef MODULE

MODULE_AUTHOR("Yen-Chau Ngo");
MODULE_DESCRIPTION("NetApp Agent driver");

int init_module(void)
{
	return agentdrv_init();
}

int cleanup_module(void)
{
	return agentdrv_cleanup();
}

#endif /* MODULE */







