/* Copyright 2004 NetApp, Inc., contribution by Eric Gumtow.
 * Copyright 2008 NetApp, Inc., contribution by Ganesh R Narayanan.
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 * License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License v2 for more
 * details.
 */ 

#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/time.h>
#include <linux/slab.h>
#include <asm/irq.h>
#include <asm-ppc/processor.h>

#include "ecc_pvt.h"

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

#ifdef MODULE
MODULE_AUTHOR("Eric Gumtow");
#endif


typedef struct {
	u_int8_t cntr;
	int overflow;
} stat8_t;

typedef struct {
	u_int16_t cntr;
	int overflow;
} stat16_t;

typedef struct {
	u_int32_t cntr;
	int overflow;
} stat32_t;

typedef struct {
	u_int64_t cntr;
	int overflow;
} stat64_t;

#define STAT_INC(x)                                                \
	if ((stats.x.cntr += 0) > (stats.x.cntr += 1)) {           \
		stats.x.cntr--;                                    \
		stats.x.overflow = 1;                              \
	}

#define STAT_MAX(x, n)                                             \
	if ((n) > stats.x.cntr) {                                  \
		stats.x.cntr = (n);                                \
	}

#define STAT_SET(x, n)                                             \
	stats.x.cntr = n

#define STAT_CLR(x)                                                \
		stats.x.cntr = 0

#define STAT_VAL(x)                                                \
		stats.x.cntr


#define MAX_LOG_ECC_ERRORS 16
#define SANE_TIME(x) ((x) > (((time_t)2004-1970)*365*24*60*60))

typedef struct {
	stat16_t total_correctable_errors;
	stat8_t total_uncorrectable_errors;
	stat64_t error_times[MAX_LOG_ECC_ERRORS];
	stat8_t previous_hr_errors;
	stat8_t previous_day_errors;
	stat8_t previous_wk_errors;
	stat8_t previous_hr_most_errors;
	stat8_t previous_day_most_errors;
	stat8_t previous_wk_most_errors;
} ecc_stats_t;



static ecc_stats_t stats;


void
ecc_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	extern void *rlm_proc_stat_buf_loc(char *);
	extern struct timeval xtime;

	int i;
	static int curr = 0;
	time_t now, hr_ago, day_ago, wk_ago;
	ecc_stats_t *ecc_buf;
	u_int32_t ecc_status;

	/* read the ECC status and clear the interrupt cause */
	mtdcr(MEMCFGA, MEM_ECCCERR);
	ecc_status = mfdcr(MEMCFGD);
	mtdcr(MEMCFGA, MEM_ECCCERR);
	mtdcr(MEMCFGD, REG_BIT(ECC_STAT_CE) | REG_BIT(ECC_STAT_UE));

	/* ordinarily we'd let the rlmprocfs module do the stats for us.
	 * but because this handler is already in the kernel, we'll take
	 * care of the stats ourselves.  here we retrieve the stat buffer 
	 * via an exported function. */
	ecc_buf = (ecc_stats_t *) rlm_proc_stat_buf_loc("/proc/rlm/ecc");
	if (!ecc_buf) {
		return;
	}

	if (ecc_status & REG_BIT(ECC_STAT_CE)) {
		STAT_INC(total_correctable_errors);
	}
	if (ecc_status & REG_BIT(ECC_STAT_UE)) {
		STAT_INC(total_uncorrectable_errors);
	}

	/* hopefully the system clock has a sane value.  if not, then
	 * we have no reliable context for when an ECC error happened.
	 * in that event, we'll stat that it happened, but not record
	 * the time of its happening. */
	now = xtime.tv_sec;
	if (!SANE_TIME(now)) {
		memcpy(ecc_buf, &stats, 320);
		return;
	}

	hr_ago = now - (1*60*60);
	day_ago = now - (24*60*60);
	wk_ago = now - (7*24*60*60);

	STAT_SET(error_times[curr], now);
	curr = (curr + 1) % MAX_LOG_ECC_ERRORS;

	STAT_CLR(previous_hr_errors);
	STAT_CLR(previous_day_errors);
	STAT_CLR(previous_wk_errors);

	for (i = 0; i < MAX_LOG_ECC_ERRORS; i++) {
		if (SANE_TIME(STAT_VAL(error_times[i]))) {
			if (STAT_VAL(error_times[i]) > wk_ago) {
				STAT_INC(previous_wk_errors);
				if (STAT_VAL(error_times[i]) > day_ago) {
					STAT_INC(previous_day_errors);
					if (STAT_VAL(error_times[i]) > hr_ago) {
						STAT_INC(previous_hr_errors);
					}
				}
			}
		}
	}

	STAT_MAX(previous_hr_most_errors, STAT_VAL(previous_hr_errors));
	STAT_MAX(previous_day_most_errors, STAT_VAL(previous_day_errors));
	STAT_MAX(previous_wk_most_errors, STAT_VAL(previous_wk_errors));

	memcpy(ecc_buf, &stats, sizeof(ecc_stats_t));
}

static int __init
ecc_init(void)
{
	int code;

	code = request_irq(RLM_ECC_IRQ, ecc_handler, SA_INTERRUPT, ECC_NAME, NULL);
	if (code) {
		printk(KERN_INFO "%s %s failed initialization: %d\n", 
		       MODULE_NAME, MODULE_VERSION, code);
		return code;
	}

	printk(KERN_INFO "%s %s initialized\n", MODULE_NAME, MODULE_VERSION);

	return 0;
}

static void __exit
ecc_exit(void)
{
	free_irq(RLM_ECC_IRQ, NULL);

	printk(KERN_INFO "%s %s removed\n", MODULE_NAME, MODULE_VERSION);
}


module_init(ecc_init);
module_exit(ecc_exit);

MODULE_DESCRIPTION(MODULE_NAME);
